import enum
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from typing import Any

from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from clios.core.main_parser import ParserAbc, ParserError
from clios.core.operator import OperatorError
from clios.core.operator_fn import OperatorFns


class _KnownTypes(enum.Enum):
    INT = int
    FLOAT = float
    TEXT = str
    BOOL = bool
    DATETIME = datetime
    DATE = date
    TIME = time
    TIMDELTA = timedelta

    @classmethod
    def _missing_(cls, value: Any) -> Any:
        return cls.TEXT


@dataclass(frozen=True)
class CliPresenter:
    operator_fns: OperatorFns
    parser: ParserAbc

    def process_error(self, error: ParserError, args: list[str]) -> None:
        """
        Process an error and present it nicely to the terminal.
        """
        # Get the error message
        message = error.message

        console = Console()

        # From context, get token_index if available
        token_index = error.ctx.get("token_index", None)
        unchainable_token_index = error.ctx.get("unchainable_token_index", None)
        error.ctx.get("num_extra_tokens", None)
        sub_error = error.ctx.get("error", None)

        if sub_error is not None:
            message += f"\n{str(sub_error)}"

        error_argument_text = Text()
        if token_index is not None:
            token_before = token_index - 1
            token_after = token_index + 1

            if token_before > 0:
                error_argument_text.append(" ")
                error_argument_text.append("...", style="dim")

            if token_before >= 0:
                error_argument_text.append(" ")
                error_argument_text.append(args[token_before], style="dim")

            error_argument_text.append(" ")
            error_argument_text.append(args[token_index], style="bold red underline")

            if token_after < len(args) and unchainable_token_index is None:
                error_argument_text.append(" ")
                error_argument_text.append(args[token_after], style="dim")

            if token_after < len(args) - 1:
                error_argument_text.append(" ")
                error_argument_text.append("...", style="dim")

            if unchainable_token_index is not None:
                error_argument_text.append(" ")
                error_argument_text.append(
                    args[unchainable_token_index], style="bold red underline"
                )

                if unchainable_token_index < len(args) - 1:
                    error_argument_text.append(" ")
                    error_argument_text.append("...", style="dim")

            console.print()
            console.print(error_argument_text)

        console.print()
        console.print(Text(message, style="bold red"))

    def print_list(self):
        """
        Prints a 2-column table dynamically adjusted to the terminal width.

        Args:
            data (list[tuple[str, str]]): A list of tuples containing data for the table.
        """
        console = Console()
        table = Table(
            show_header=True, header_style="bold blue", title="Available Operators"
        )
        data = [(name, op.short_description) for name, op in self.operator_fns.items()]
        # Add columns with dynamic width for the second column
        if not data:
            console.print("No operators found!", style="bold red")
            return
        column1_width = max(len(row[0]) for row in data)
        column1_width = max(len("Operator"), column1_width)
        table.add_column(
            "Operator", no_wrap=True, min_width=column1_width, style="bold"
        )
        # table.add_column("Description", width=console.size.width - column1_width - 10)
        table.add_column("Description")
        # Add rows
        for col1, col2 in data:
            table.add_row(col1, col2)
        console.print()
        console.print(table)

    def print_detail(self, name: str, exe_name: str = ""):
        """
        Prints a detailed operator page using Rich.
        """
        console = Console()
        if self.parser.is_inline_operator_name(name):
            try:
                op_fn = self.parser.get_inline_operator_fn(name, 0)
            except ParserError as e:
                console.print(e.message, style="bold red")
                raise SystemExit(1)
        else:
            try:
                op_fn = self.operator_fns[name]
            except KeyError:
                console.print(f"Operator `{name}` not found!", style="bold red")
                raise SystemExit(1)

        synopsis = self.parser.get_synopsis(op_fn, name)

        if exe_name:
            synopsis = f"{exe_name}{synopsis}"

        short_description = op_fn.short_description
        long_description = op_fn.long_description
        args_doc: list[dict[str, str]] = []
        kwds_doc: list[dict[str, str]] = []
        for param in op_fn.parameters:
            if param.is_positional_param or param.is_var_param:
                docs = args_doc
            elif param.is_keyword_param or param.is_var_keyword:
                docs = kwds_doc
            else:
                continue
            doc: dict[str, Any] = {}
            doc["name"] = param.name
            doc["type"] = _KnownTypes(param.type_).name
            doc["required"] = "Required" if param.is_required else "Optional"
            doc["default_value"] = str(param.default)
            if param.default == "":
                doc["default_value"] = "''"
            if param.default is param.empty:
                doc["default_value"] = ""
            doc["description"] = param.description
            doc["choices"] = ""
            if param.choices:
                doc["choices"] = ", ".join(param.choices)
            docs.append(doc)

        # Synopsis
        synopsis_panel = Panel(
            synopsis,
            title="Synopsis",
            style="bold yellow",
            title_align="left",
            padding=(1, 2),
        )

        # Render all parts
        console.print(synopsis_panel)
        description_text = Text()
        if short_description:
            description_text.append(f"{short_description}\n\n", style="bold blue")
        if long_description:
            description_text.append(long_description, style="dim")
        if short_description or long_description:
            description_panel = Panel(
                description_text,
                title="Description",
                style="bold blue",
                title_align="left",
                padding=(1, 2),
            )
            console.print(description_panel)

        if args_doc:
            param_table_args = _create_param_table(
                args_doc, title="Positional Arguments"
            )
            console.print(param_table_args)

        if kwds_doc:
            param_table_kwds = _create_param_table(kwds_doc, title="Keyword Arguments")
            console.print(param_table_kwds)

        examples: list[str] = []
        for _, example in op_fn.examples:
            examples.append(example)

        if examples:
            text = "\n".join(examples)
            examples_text = Text()
            examples_text.append(text, style="italic")
            examples_panel = Panel(
                examples_text,
                title="Examples",
                style="bold magenta",
                title_align="left",
                padding=(1, 2),
            )
            console.print(examples_panel)

    def dry_run(self, args: list[str]):
        """
        Dry run the operator function with the given arguments.
        """
        console = Console()
        try:
            operator = self.parser.get_operator(self.operator_fns, args)
        except ParserError as e:
            self.process_error(e, args)
            raise SystemExit(1)
        console.print(operator.draw())

    def run(self, args: list[str], debug: bool = False):
        """
        Run the operator function with the given arguments.
        """
        try:
            operator = self.parser.get_operator(self.operator_fns, args)
        except ParserError as e:
            self.process_error(e, args)
            if debug:
                raise e
            raise SystemExit(1)

        try:
            operator.execute()
        except OperatorError as e:
            Console().print(Text(str(e), style="bold red"))
            if debug:
                raise e
            raise SystemExit(1)


def _create_param_table(args_doc: list[dict[str, str]], title: str) -> Table:
    param_table = Table(title=title, show_header=True, header_style="bold magenta")
    param_table.add_column("Parameter", style="dim", no_wrap=True)
    param_table.add_column("Type", style="dim", no_wrap=True)
    param_table.add_column("Required", style="dim", no_wrap=True)
    param_table.add_column("Description", justify="left")
    for doc in args_doc:
        description = doc["description"]
        default_value = doc["default_value"]
        choices = doc["choices"]
        if choices != "":
            description += f" Choices: {choices}"
        if default_value != "":
            description += f" (default: {default_value})"
        param_table.add_row(
            doc["name"],
            doc["type"],
            doc["required"],
            description,
        )
    return param_table



import typing as t

from rich import print

from .cli.app import Clios as Clios
from .cli.app import OperatorFns as OperatorFns
from .cli.app import operator as operator
from .core.exceptions import CliosError as CliosError
from .core.operator import OperatorError as OperatorError
from .core.param_info import Input as Input
from .core.param_info import Output as Output
from .core.param_info import Param as Param

registry = OperatorFns()


@registry.register(name="print")
def _output(input: t.Annotated[t.Any, Input()]) -> None:
    """
    Print the given input data to the terminal.

    description:
        It uses the `rich` library to print the data in a formatted way.
    """
    print(input)  # noqa: T201


main = Clios(registry)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable

from pydantic import ValidationError

from .exceptions import CliosError
from .operator_fn import OperatorFn


class OperatorError(Exception):
    def __init__(self, message: str, ctx: dict[str, Any] = {}) -> None:
        self.message = message
        self.ctx = ctx
        super().__init__(self.message)

    def __str__(self) -> str:
        message = self.message
        if "error" in self.ctx:
            message += f"\n{self.ctx['error']}"
        return message


@dataclass(frozen=True)
class OperatorAbc(ABC):
    @abstractmethod
    def execute(self) -> Any: ...  # pragma: no cover

    @abstractmethod
    def draw(self) -> str: ...  # pragma: no cover


@dataclass(frozen=True)
class SimpleOperator(OperatorAbc):
    """
    An operator that take an input and returns the input as output

    This is used to to represent an input parameter as an operator
    """

    name: str
    index: int
    input_: Any

    def execute(self) -> Any:
        return self.input_

    def draw(self) -> str:
        return self.name


@dataclass(frozen=True)
class BaseOperator(OperatorAbc):
    """A base class for operators"""

    name: str
    index: int
    operator_fn: OperatorFn
    args: tuple[Any, ...] = ()
    kwds: tuple[tuple[str, Any], ...] = ()

    def _validate_arguments(self) -> list[Any]:
        arg_values: list[Any] = []
        iter_args = self.operator_fn.parameters.iter_positional_arguments()
        for val in self.args:
            param = next(iter_args)
            try:
                arg_values.append(param.execute_phase_validator.validate_python(val))
            except ValidationError as e:
                raise OperatorError(
                    f"Data validation failed for the argument `{param.name}` of operator `{self.name}`!",
                    ctx={"error": e, "index": self.index, "name": self.name},
                )
        return arg_values

    def _validate_keywords(self) -> dict[str, Any]:
        arg_values: dict[str, Any] = {}
        for key, val in self.kwds:
            param = self.operator_fn.parameters.get_keyword_argument(key)
            try:
                arg_values[key] = param.execute_phase_validator.validate_python(val)
            except ValidationError as e:
                raise OperatorError(
                    f"Data validation failed for the argument `{param.name}` of operator `{self.name}`!",
                    ctx={"error": e, "index": self.index, "name": self.name},
                )
        return arg_values

    def _compose_arg_values(self, args: list[Any], inputs: list[Any]) -> list[Any]:
        iter_params = self.operator_fn.parameters.iter_positional()
        args_rev = list(args[::-1])
        inputs_rev = list(inputs[::-1])
        param_values: list[Any] = []
        for param in iter_params:
            if len(args_rev) == 0 and len(inputs_rev) == 0:
                break
            if param.is_input:
                param_values.append(inputs_rev.pop())
            else:
                param_values.append(args_rev.pop())
        return param_values

    def execute(self) -> Any:
        arg_values = self._validate_arguments()
        kwds_values = self._validate_keywords()
        try:
            value = self.operator_fn.callback(*arg_values, **kwds_values)
        except CliosError as e:
            raise OperatorError(
                f"An error occurred while executing operator `{self.name}`!",
                ctx={"error": e, "index": self.index, "name": self.name},
            )

        try:
            return self.operator_fn.output.validator.validate_python(value)
        except ValidationError as e:
            raise OperatorError(
                f"Data validation failed for the output of operator `{self.name}`!\n"
                + f"Error: {e}\n"
                + "It's a bug! Please report it!",
                ctx={"index": self.index, "name": self.name},
            )

    def draw(self) -> str:
        return f"{self.name}"


@dataclass(frozen=True)
class LeafOperator(BaseOperator):
    """An operator that has no inputs"""

    pass


@dataclass(frozen=True)
class _Operator(BaseOperator):
    inputs: tuple[Any, ...] = ()

    def execute_input(self, input_: Any) -> Any:
        raise NotImplementedError

    def _validate_execute_inputs(self) -> list[Any]:
        input_values: list[Any] = []
        iter_inputs = self.operator_fn.parameters.iter_inputs()
        for input_ in self.inputs:
            input_param = next(iter_inputs)
            try:
                value = input_param.execute_phase_validator.validate_python(
                    self.execute_input(input_)
                )
            except ValidationError as e:
                raise OperatorError(
                    f"Data validation failed for the input of operator `{self.name}`!",
                    ctx={"error": e, "index": self.index, "name": self.name},
                )
            input_values.append(value)
        return input_values

    def execute(self) -> Any:
        arg_values = self._validate_arguments()
        kwds_values = self._validate_keywords()
        input_values = self._validate_execute_inputs()
        positional_args = self._compose_arg_values(arg_values, input_values)

        try:
            value = self.operator_fn.callback(*positional_args, **kwds_values)
        except CliosError as e:
            raise OperatorError(
                f"An error occurred while executing operator `{self.name}`!",
                ctx={"error": e, "index": self.index, "name": self.name},
            )

        try:
            return self.operator_fn.output.validator.validate_python(value)
        except ValidationError as e:
            raise OperatorError(
                f"Data validation failed for the output of operator `{self.name}`!\n"
                + f"Error: {e}\n"
                + "It's a bug! Please report it!",
                ctx={"index": self.index, "name": self.name},
            )


@dataclass(frozen=True)
class Operator(_Operator):
    """An operator that has inputs and the inputs are also operators"""

    inputs: tuple[OperatorAbc, ...] = ()

    def execute_input(self, input_: OperatorAbc) -> Any:
        return input_.execute()

    def draw(self) -> str:
        res = f"{self.name} [ "
        for input_ in self.inputs:
            res += f"{input_.draw()} "
        res += "]"
        return res


@dataclass(frozen=True)
class DelegateOperator(_Operator):
    """An operator that delegates the string inputs as it is"""

    inputs: tuple[str, ...] = ()

    def execute_input(self, input_: str) -> str:
        return input_

    def draw(self) -> str:
        res = f"{self.name} [ "
        if self.inputs:
            res += " ".join(self.inputs)
            res += " "
        res += "]"
        return res


@dataclass(frozen=True)
class RootOperator(OperatorAbc):
    input: BaseOperator
    callback: Callable[..., Any]
    args: tuple[str, ...] = ()

    def execute(self) -> Any:
        value = self.input.execute()
        return self.callback(value, *self.args)

    def draw(self) -> str:
        res = ",".join(self.args)
        if res:
            res = f"{res} "
        res += f"[ {self.input.draw()} ]"
        return res



import typing as t

import click
from rich import print

from clios.core.param_parser import ParamParserAbc

from ..core.operator_fn import OperatorFn
from ..core.operator_fn import OperatorFns as OperatorFns_
from .main_parser import CliParser
from .param_parser import StandardParamParser
from .presenter import CliPresenter

standard_param_parser = StandardParamParser()


def operator(
    *,
    param_parser: ParamParserAbc = standard_param_parser,
    implicit: t.Literal["input", "param"] = "param",
) -> t.Callable[..., t.Any]:
    def decorator(func: t.Callable[..., t.Any]) -> OperatorFn:
        operator_fn = OperatorFn.from_def(
            func, param_parser=param_parser, implicit=implicit
        )
        return operator_fn

    return decorator


class OperatorFns(OperatorFns_):
    @t.override
    def register(
        self,
        *,
        name: str = "",
        param_parser: ParamParserAbc = standard_param_parser,
        implicit: t.Literal["input", "param"] = "param",
        is_delegate: bool = False,
    ) -> t.Callable[..., t.Any]:
        return super().register(
            name=name,
            param_parser=param_parser,
            implicit=implicit,
            is_delegate=is_delegate,
        )


class Clios:
    def __init__(self, operator_fns: OperatorFns_, exe_name: str = "") -> None:
        self._operators = operator_fns
        self._parser = CliParser()
        self._presenter = CliPresenter(self._operators, self._parser)
        self._exe_name = exe_name

    def __call__(self):
        try:
            res = _click_app(standalone_mode=False)
        except click.exceptions.UsageError as e:
            print(e.format_message())
            with click.Context(_click_app) as ctx:
                click.echo(_click_app.get_help(ctx))
            raise SystemExit(1)

        if isinstance(res, tuple):
            args, options = res
        elif res != 0:
            raise SystemExit(res)
        else:
            return

        if options["list"]:
            return self._presenter.print_list()
        if options["show"] is not None:
            return self._presenter.print_detail(options["show"], self._exe_name)
        if options["dry_run"]:
            return self._presenter.dry_run(args)
        debug = options["debug"]
        return self._presenter.run(args, debug)


@click.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
@click.option("--list", type=bool, help="List all available operators", is_flag=True)
@click.option("--debug", type=bool, help="Turn on debugging", is_flag=True)
@click.option(
    "--show", type=str, help="Show the help information for the given operator", nargs=1
)
@click.option(
    "--dry-run", type=bool, help="Dry run: prints the call tree", is_flag=True
)
@click.pass_context
def _click_app(ctx: t.Any, **kwargs: t.Any) -> tuple[list[str], dict[str, t.Any]]:
    return ctx.args, kwargs



import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from clios.core.tokenizer import Token, Tokenizer


@dataclass(frozen=True)
class StringToken(Token):
    """A string token will match any string"""

    @classmethod
    def match(cls, string: str) -> bool:
        return True


@dataclass(frozen=True)
class OperatorToken(Token):
    """An operator token will match any string starting with '-'"""

    @classmethod
    def match(cls, string: str) -> bool:
        if string[0] != "-":
            return False
        name = Path(string[1:]).name
        pattern = r"^[a-zA-Z_]"
        if re.match(pattern, name) is not None:
            return True
        return False


@dataclass(frozen=True)
class LeftBracketToken(Token):
    value: str = "["

    @classmethod
    def match(cls, string: str) -> bool:
        return string == "["


@dataclass(frozen=True)
class RightBracketToken(Token):
    value: str = "]"

    @classmethod
    def match(cls, string: str) -> bool:
        return string == "]"


@dataclass(frozen=True)
class ColonToken(Token):
    value: str = ":"

    @classmethod
    def match(cls, string: str) -> bool:
        return string == ":"


class TokenType(Enum):
    COLON = ColonToken
    LEFT_BRACKET = LeftBracketToken
    RIGHT_BRACKET = RightBracketToken
    OPERATOR = OperatorToken
    STRING = StringToken

    @classmethod
    def _missing_(cls, value: object) -> "TokenType":
        for member in cls:
            if member.value.match(value) and member is not cls.STRING:
                return member
        # if no match, return STRING
        return cls.STRING


class CliTokenizer(Tokenizer[list[str]]):
    def tokenize(self, input: list[str]) -> tuple[Token, ...]:
        tokens: list[Token] = []
        for arg in input:
            token = TokenType(arg)
            tokens.append(token.value(arg))
        return tuple(tokens)



# from dataclasses import dataclass
import enum
import inspect as i
import typing as t
from dataclasses import dataclass
from enum import Enum
from functools import cached_property

from pydantic import BeforeValidator, ConfigDict, PydanticUserError, Strict, TypeAdapter
from pydantic.functional_validators import AfterValidator, PlainValidator, WrapValidator
from typing_extensions import Doc

from .param_info import Input, Output, Param, ParamTypes
from .utils import get_parameter_type_annotation


def _get_type(annotation: t.Any):
    if t.get_origin(annotation) is t.Annotated:
        return t.get_args(annotation)[0]
    return annotation


def _init_validator(
    info: ParamTypes,
    annotation: t.Any,
    phase_validators: tuple[t.Callable[[t.Any], t.Any], ...],
    phase: str,
) -> TypeAdapter[t.Any]:
    validators = [BeforeValidator(validator) for validator in phase_validators]
    if info.core_validation_phase != phase:
        annotation = t.Any

    if validators:
        annotation = t.Annotated[annotation, *validators]  # type: ignore

    if info.strict and _get_type(annotation) is not t.Any:
        annotation = t.Annotated[annotation, Strict()]

    try:
        return TypeAdapter(annotation, config=ConfigDict(arbitrary_types_allowed=True))
    except PydanticUserError as e:
        if e.code == "type-adapter-config-unused":
            return TypeAdapter(annotation)
        raise


def _get_description(annotation: t.Any) -> str:
    if t.get_origin(annotation) is t.Annotated:
        for arg in t.get_args(annotation)[::-1]:
            if isinstance(arg, Doc):
                return arg.documentation
    return ""


class ParameterKind(Enum):
    # POSITIONAL_OR_KEYWORD is as POSITIONAL_ONLY
    POSITIONAL_ONLY = i.Parameter.POSITIONAL_ONLY
    VAR_POSITIONAL = i.Parameter.VAR_POSITIONAL
    KEYWORD_ONLY = i.Parameter.KEYWORD_ONLY
    VAR_KEYWORD = i.Parameter.VAR_KEYWORD

    @classmethod
    def _missing_(cls, value: object) -> "ParameterKind":
        if value == i.Parameter.POSITIONAL_OR_KEYWORD:
            return cls.POSITIONAL_ONLY
        return super()._missing_(value)


@dataclass(frozen=True)
class _Empty:
    pass


@dataclass
class ParamVal:
    val: t.Any
    key: str = ""


@dataclass(frozen=True)
class Parameter:
    """A dataclass to represent a parameter of an operator function"""

    empty: t.ClassVar[_Empty] = _Empty()
    name: str
    kind: ParameterKind
    info: ParamTypes
    annotation: t.Any
    build_phase_validator: TypeAdapter[t.Any]
    execute_phase_validator: TypeAdapter[t.Any]
    description: str
    default: t.Any

    @property
    def choices(self) -> list[t.Any]:
        if t.get_origin(_get_type(self.annotation)) is t.Literal:
            return list(t.get_args(_get_type(self.annotation)))
        if issubclass(_get_type(self.annotation), enum.Enum):
            return [choice.value for choice in _get_type(self.annotation)]
        return []

    @classmethod
    def validate(cls, param: i.Parameter, implicit: str) -> "Parameter":
        assert param.annotation is not i.Signature.empty, (
            f"Missing type annotation for parameter `{param.name}`"
        )
        type_ = (
            t.get_args(param.annotation)[0]
            if t.get_origin(param.annotation) is t.Annotated
            else param.annotation
        )

        assert not _is_not_supported(type_), (
            f"Unsupported type annotation for parameter `{param.name}`"
        )

        default = param.default
        if param.default is i.Signature.empty:
            default = Parameter.empty

        param_type = get_parameter_type_annotation(param)
        if param_type is None:
            if implicit == "input":
                param_type = Input()
            elif implicit == "param":
                param_type = Param()
            else:
                raise AssertionError(f"Invalid implicit option `{implicit}`")
        kind = ParameterKind(param.kind)

        if isinstance(param_type, Input):
            assert kind not in (
                ParameterKind.VAR_KEYWORD,
                ParameterKind.KEYWORD_ONLY,
            ), f"Input parameter `{param.name}` cannot be keyword argument"

            assert default is Parameter.empty, (
                f"Input parameter `{param.name}` cannot have a default value"
            )

        build_phase_validator = _init_validator(
            param_type,
            param.annotation,
            param_type.build_phase_validators,
            "build",
        )
        execute_phase_validator = _init_validator(
            param_type,
            param.annotation,
            param_type.execute_phase_validators,
            "execute",
        )

        description = _get_description(param.annotation)
        return cls(
            name=param.name,
            kind=kind,
            info=param_type,
            annotation=param.annotation,
            build_phase_validator=build_phase_validator,
            execute_phase_validator=execute_phase_validator,
            description=description,
            default=default,
        )

    @property
    def is_input(self):
        """Check if the parameter is an input"""
        return isinstance(self.info, Input)

    @property
    def is_param(self):
        """Check if the parameter is a parameter"""
        return isinstance(self.info, Param)

    @property
    def type_(self):
        """Get the data type of the parameter"""
        return _get_type(self.annotation)

    @property
    def is_var_keyword(self):
        """Check if the parameter is a variable keyword"""
        return self.kind == ParameterKind.VAR_KEYWORD

    @property
    def is_var_input(self):
        """Check if the parameter is a variable input"""
        return self.is_input and self.kind == ParameterKind.VAR_POSITIONAL

    @property
    def is_var_param(self):
        """Check if the parameter is a variable parameter"""
        return self.is_param and self.kind == ParameterKind.VAR_POSITIONAL

    @property
    def is_keyword_param(self):
        """Check if the parameter is a keyword parameter"""
        return self.kind == ParameterKind.KEYWORD_ONLY and not self.is_input

    @property
    def is_positional_param(self):
        """Check if the parameter is a positional parameter"""
        return self.kind == ParameterKind.POSITIONAL_ONLY and not self.is_input

    @property
    def is_required(self):
        """Check if the parameter is required"""
        if self.is_input:
            return True

        if self.is_var_param or self.is_var_keyword:
            return False

        return self.default is Parameter.empty


@dataclass(frozen=True)
class ReturnValue:
    """A dataclass to represent the return type of an operator function"""

    validator: TypeAdapter[t.Any]
    annotation: t.Any
    info: Output

    @classmethod
    def validate(cls, annotation: t.Any, info: Output) -> "ReturnValue":
        prohibited_validators = (PlainValidator, WrapValidator, AfterValidator)
        type_adapter: TypeAdapter[t.Any]
        if _get_type(annotation) is None:
            type_adapter = TypeAdapter(None)
            info = Output(callback=info.callback, num_outputs=0)
        else:
            if t.get_origin(annotation) is t.Annotated:
                if any(
                    isinstance(arg, prohibited_validators)
                    for arg in t.get_args(annotation)[1:]
                ):
                    raise AssertionError(
                        "Only `BeforeValidator` is allowed on return value!"
                    )
            if info.strict:
                annotation = t.Annotated[annotation, Strict()]
            type_adapter = TypeAdapter(
                annotation, config=ConfigDict(arbitrary_types_allowed=True)
            )
        return cls(type_adapter, annotation, info)

    @property
    def type_(self):
        return _get_type(self.annotation)


class Parameters(tuple[Parameter, ...]):
    @cached_property
    def required_keywords(self) -> dict[str, Parameter]:
        """Get a dictionary of required keyword arguments"""
        return {
            param.name: param
            for param in self
            if param.is_keyword_param and param.is_required
        }

    def iter_positional_arguments(self) -> t.Generator[Parameter, None, None]:
        """Iterate over positional arguments"""
        for param in self:
            if param.is_positional_param:
                yield param
        while self.var_argument is not None:
            yield self.var_argument

    def iter_inputs(self) -> t.Generator[Parameter, None, None]:
        """Iterate over input arguments"""
        for param in self:
            if param.is_input and not param.is_var_input:
                yield param
        while self.var_input is not None:
            yield self.var_input

    def iter_positional(self) -> t.Generator[Parameter, None, None]:
        """Iterate over positional (i.e. both input and arguments)"""
        var_param = None
        for param in self:
            if param.is_var_param or param.is_var_input:
                var_param = param
                break
            if param.is_keyword_param or param.is_var_keyword:
                break
            yield param

        while var_param is not None:
            yield var_param

    def get_keyword_argument(self, key: str) -> Parameter:
        """Get the keyword argument by key"""
        for param in self:
            if param.is_keyword_param:
                if param.name == key:
                    return param
        if self.var_keyword is not None:
            return self.var_keyword
        raise KeyError(f"Unknown keyword argument: {key}")

    @cached_property
    def var_argument(self) -> Parameter | None:
        """Get the variable positional argument, if any"""
        for param in self:
            if param.is_var_param:
                return param
        return None

    @cached_property
    def var_keyword(self) -> Parameter | None:
        """Get the variable keyword argument, if any"""
        for param in self:
            if param.is_var_keyword:
                return param
        return None

    @cached_property
    def num_minimum_inputs(self) -> int:
        """Get the number of minimum required input parameters"""
        return len([param for param in self if param.is_input])

    @cached_property
    def input_present(self) -> int:
        """Check if there is any input parameter"""
        for param in self:
            if param.is_input:
                return True
        return False

    @cached_property
    def var_input(self) -> Parameter | None:
        """Get the variable input parameter, if any"""
        for param in self:
            if param.is_var_input:
                return param
        return None

    @cached_property
    def num_required_args(self) -> int:
        """Get the number of required positional arguments"""
        params = [
            param for param in self if param.is_required and param.is_positional_param
        ]
        return len(params)


def _is_not_supported(type_: t.Any) -> bool:
    _builtin_generic_types = [  # type: ignore
        list,
        dict,
        set,
        tuple,
        frozenset,
    ]
    if t.get_origin(type_) is t.Literal:
        return False
    return not isinstance(type_, type) or type_ in _builtin_generic_types



import importlib.util
import logging
import re
import typing as t
from dataclasses import dataclass
from pathlib import Path

from pydantic import ValidationError

from clios.core.main_parser import ParserAbc, ParserError
from clios.core.operator import (
    BaseOperator,
    DelegateOperator,
    LeafOperator,
    Operator,
    OperatorAbc,
    RootOperator,
    SimpleOperator,
)
from clios.core.operator_fn import OperatorFn, OperatorFns
from clios.core.param_parser import ParamParserError
from clios.core.tokenizer import Token

from .tokenizer import (
    CliTokenizer,
    LeftBracketToken,
    OperatorToken,
    RightBracketToken,
    StringToken,
    Tokenizer,
)

logger = logging.getLogger(__name__)


def simple_callback(value: t.Any) -> t.Any:
    return value


@dataclass(frozen=True)
class CliParser(ParserAbc):
    tokenizer: Tokenizer[list[str]] = CliTokenizer()

    def get_name(self, token: Token) -> str:
        if isinstance(token, OperatorToken):
            return token.value.split(",")[0].lstrip("-")
        return token.value

    def get_param_string(self, token: Token) -> str:
        if isinstance(token, OperatorToken):
            return ",".join(token.value.split(",")[1:])
        return ""

    def is_inline_operator(self, token: Token) -> bool:
        if isinstance(token, OperatorToken):
            name = Path(self.get_name(token)).name
            return self.is_inline_operator_name(name)
        return False

    def is_inline_operator_name(self, name: str) -> bool:
        pattern = r"^[a-zA-Z_][a-zA-Z0-9_]*\.py$"
        return re.match(pattern, name) is not None

    def get_operator(
        self,
        operator_fns: OperatorFns,
        input: list[str],
        callback: t.Callable[..., t.Any] = simple_callback,
        **kwargs: t.Any,
    ) -> RootOperator:
        if not input:
            raise ParserError("Input is empty!")
        tokens = list(self.tokenizer.tokenize(input))
        num_tokens = len(tokens)
        token = tokens.pop(0)
        operator_name = self.get_name(token)
        param_string = self.get_param_string(token)
        operator_fn = self._get_operator_fn(operator_fns, token, 0)

        num_outputs = 0
        if operator_fn.output.type_ is not None:
            if operator_fn.output.info.callback is None:
                raise ParserError(
                    f"Operator `{operator_name}` cannot be used as root operator!",
                    ctx={"token_index": 0},
                )
            callback = operator_fn.output.info.callback
            num_outputs = operator_fn.output.info.num_outputs

        output_file_paths: list[str] = []
        for i in range(num_outputs):
            try:
                output_token = tokens.pop()
                if not isinstance(output_token, StringToken):
                    raise ParserError(
                        "Output file path must be a string",
                        ctx={"token_index": num_tokens - 1 - i},
                    )
                output_file_paths.append(str(output_token.value))
            except IndexError:
                raise ParserError(
                    "Missing output(s)!", ctx={"token_index": num_tokens - 1 - i}
                )

        tokens.reverse()

        operator = self._get_operator(
            operator_fns=operator_fns,
            operator_name=operator_name,
            param_string=param_string,
            operator_fn=operator_fn,
            token_index=0,
            input_tokens=tokens,
        )

        if len(tokens) != 0:
            raise ParserError(
                "Got too many inputs!",
                ctx={
                    "num_extra_tokens": len(tokens),
                    "token_index": num_tokens - len(tokens),
                },
            )

        return RootOperator(
            input=operator,
            callback=callback,
            args=tuple(output_file_paths),
        )

    def _get_operator(
        self,
        operator_fns: OperatorFns,
        operator_name: str,
        param_string: str,
        operator_fn: OperatorFn,
        token_index: int,
        input_tokens: list[Token],
    ) -> BaseOperator:
        try:
            logger.debug(
                f"parsing arguments for operator `{operator_name}`, param_string: {param_string}"
            )
            args, kwds = operator_fn.param_parser.parse_arguments(
                string=param_string,
                parameters=operator_fn.parameters,
            )
        except ParamParserError as e:
            raise ParserError(
                f"Failed to parse arguments for operator `{operator_name}`!",
                ctx={"error": e, "token_index": token_index},
            )

        if not operator_fn.parameters.input_present:
            return LeafOperator(
                name=operator_name,
                index=token_index,
                operator_fn=operator_fn,
                args=args,
                kwds=kwds,
            )

        if not input_tokens:
            raise ParserError(
                f"Missing inputs for operator {operator_name}!",
                ctx={"token_index": token_index},
            )
        operator_token_index = token_index
        num_tokens = len(input_tokens)
        if isinstance(input_tokens[-1], LeftBracketToken):
            # pop the left bracket
            input_tokens.pop()
            token_index += 1
            # get the child tokens until closing right bracket
            child_tokens = self._get_bracketed_tokens(input_tokens, token_index)
        else:
            child_tokens = input_tokens

        inputs = self._parse_inputs(
            operator_fns,
            operator_fn,
            token_index,
            operator_token_index,
            child_tokens,
        )

        token_index += num_tokens - len(input_tokens)

        if len(inputs) < operator_fn.parameters.num_minimum_inputs:
            raise ParserError(
                f"Missing inputs for operator {operator_name}!",
                ctx={"token_index": operator_token_index},
            )

        if operator_fn.is_delegate:
            return DelegateOperator(
                name=operator_name,
                index=operator_token_index,
                operator_fn=operator_fn,
                inputs=tuple(inputs),
                args=args,
                kwds=kwds,
            )

        return Operator(
            name=operator_name,
            index=token_index,
            operator_fn=operator_fn,
            inputs=tuple(inputs),
            args=args,
            kwds=kwds,
        )

    def _parse_inputs(
        self,
        operator_fns: OperatorFns,
        operator_fn: OperatorFn,
        token_index: int,
        operator_token_index: int,
        input_tokens: list[Token],
    ):
        if operator_fn.is_delegate:
            return self._parse_delegate_inputs(
                operator_fns,
                operator_fn,
                token_index,
                operator_token_index,
                input_tokens,
            )
        return self._parse_input_operators(
            operator_fns, operator_fn, token_index, operator_token_index, input_tokens
        )

    def _parse_delegate_inputs(
        self,
        operator_fns: OperatorFns,
        operator_fn: OperatorFn,
        token_index: int,
        operator_token_index: int,
        input_tokens: list[Token],
    ):
        inputs: list[str] = []

        child_index = token_index
        for input_param in operator_fn.parameters.iter_inputs():
            if len(input_tokens) == 0:
                break
            child_token = input_tokens.pop()
            child_index += 1
            try:
                value = input_param.build_phase_validator.validate_python(
                    child_token.value
                )
            except ValidationError as e:
                raise ParserError(
                    f"Data validation failed for input {child_token.value}!",
                    ctx={"error": e, "token_index": child_index},
                )
            inputs.append(value)
        return inputs

    def _parse_input_operators(
        self,
        operator_fns: OperatorFns,
        operator_fn: OperatorFn,
        token_index: int,
        operator_token_index: int,
        input_tokens: list[Token],
    ):
        child_operators: list[OperatorAbc] = []

        child_index = token_index
        for input_param in operator_fn.parameters.iter_inputs():
            if len(input_tokens) == 0:
                break
            child_token = input_tokens.pop()
            child_index += 1
            if isinstance(child_token, OperatorToken):
                child_op_name = self.get_name(child_token)
                child_op_param_string = self.get_param_string(child_token)
                child_op_fn = self._get_operator_fn(
                    operator_fns, child_token, child_index
                )
                in_type = input_param.type_
                if in_type is not t.Any and in_type != child_op_fn.output.type_:
                    raise ParserError(
                        "These operators cannot be chained together!",
                        ctx={
                            "unchainable_token_index": child_index,
                            "token_index": operator_token_index,
                        },
                    )
                num_tokens_before = len(input_tokens)
                child_operator = self._get_operator(
                    operator_fns=operator_fns,
                    operator_name=child_op_name,
                    param_string=child_op_param_string,
                    operator_fn=child_op_fn,
                    token_index=child_index,
                    input_tokens=input_tokens,
                )
                child_operators.append(child_operator)
                child_index += num_tokens_before - len(input_tokens)
            elif isinstance(child_token, StringToken):
                try:
                    value = input_param.build_phase_validator.validate_python(
                        child_token.value
                    )
                except ValidationError as e:
                    raise ParserError(
                        f"Data validation failed for input {child_token.value}!",
                        ctx={"error": e, "token_index": child_index},
                    )
                child_operators.append(
                    SimpleOperator(
                        name=child_token.value,
                        index=child_index,
                        input_=value,
                    )
                )
            else:
                raise ParserError(
                    "This syntax is not supported yet!",
                    ctx={"token_index": child_index},
                )

        return child_operators

    def get_inline_operator_fn(self, name: str, index: int) -> OperatorFn:
        module_path = Path(name)
        module_name = module_path.stem
        try:
            module = load_module(module_name, module_path)
        except (ModuleNotFoundError, FileNotFoundError):
            raise ParserError(
                f"Module `{name}` not found!",
                ctx={"token_index": index},
            )
        operator_fn: OperatorFn | None = None

        for attr_name in dir(module):
            attr = getattr(module, attr_name)
            if isinstance(attr, OperatorFn):
                if operator_fn is not None:
                    raise ParserError(
                        f"Multiple OperatorFn found in module `{name}`!",
                        ctx={"token_index": index},
                    )
                operator_fn = attr
        if operator_fn is None:
            raise ParserError(
                f"No OperatorFn found in module `{name}`!",
                ctx={"token_index": index},
            )
        return operator_fn

    def get_synopsis(
        self,
        operator_fn: OperatorFn,
        operator_name: str,
        command_name="",
        **kwds: t.Any,
    ) -> str:
        op = operator_fn
        param_synopsis = op.param_parser.get_synopsis(op.parameters, lsep=",")
        input_synopsis = " ".join([i.name for i in op.parameters if i.is_input])
        output_synopsis = " ".join(
            [f"output{i + 1}" for i in range(op.output.info.num_outputs)]
        )
        if op.output.info.num_outputs == 1:
            output_synopsis = "output"

        synopsis = command_name
        synopsis += f" -{operator_name}"
        if param_synopsis:
            synopsis += param_synopsis
        if input_synopsis:
            synopsis += f" {input_synopsis}"
        if output_synopsis:
            synopsis += f" {output_synopsis}"
        return synopsis

    def _get_operator_fn(
        self,
        operator_fns: OperatorFns,
        token: Token,
        index: int,
    ) -> OperatorFn:
        name = self.get_name(token)

        if self.is_inline_operator(token):
            return self.get_inline_operator_fn(name, index)

        op = operator_fns.get(name)
        if op is None:
            raise ParserError(
                f"Operator `{name}` not found!", ctx={"token_index": index}
            )
        return op

    def _get_bracketed_tokens(
        self, tokens: list[Token], opening_index: int
    ) -> list[Token]:
        """Get tokens until closing bracket is found"""
        bracket_stack: list[Token] = []
        token_stack: list[Token] = []
        while tokens:
            token = tokens.pop()
            token_stack.append(token)
            if isinstance(token, LeftBracketToken):
                bracket_stack.append(token)
            elif isinstance(token, RightBracketToken):
                if not bracket_stack:
                    return list(reversed(token_stack[:-1]))
                bracket_stack.pop()

        raise ParserError(
            "Missing closing bracket!",
            ctx={"token_index": opening_index},
        )


def load_module(module_name, path):
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



from typing import Any, Callable, Literal

from pydantic import NonNegativeInt
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class _ParamInfoBase:
    strict: bool = False
    build_phase_validators: tuple[Callable[[Any], Any], ...] = ()
    execute_phase_validators: tuple[Callable[[Any], Any], ...] = ()
    core_validation_phase: Literal["build", "execute"] = "build"


@dataclass(frozen=True)
class Param(_ParamInfoBase):
    pass


@dataclass(frozen=True)
class Input(_ParamInfoBase):
    pass


ParamTypes = Param | Input


@dataclass(frozen=True)
class Output:
    strict: bool = False
    callback: Callable[..., None] | None = None
    num_outputs: NonNegativeInt = 1



import inspect
import typing as t

from pydantic._internal._typing_extra import eval_type_lenient as evaluate_forwardref

from .param_info import Input, Output, Param, ParamTypes


def get_output_info(return_annotation: t.Any) -> Output:
    if t.get_origin(return_annotation) is not t.Annotated:
        return Output()

    for arg in reversed(t.get_args(return_annotation)[1:]):
        if isinstance(arg, Output):
            return arg
    return Output()


def get_parameter_type_annotation(param: inspect.Parameter) -> ParamTypes | None:
    if t.get_origin(param.annotation) is t.Annotated:
        annotated_args = t.get_args(param.annotation)
        for arg in reversed(annotated_args[1:]):
            if isinstance(arg, (Param, Input)):
                return arg
    return None


def get_typed_signature(call: t.Callable[..., t.Any]) -> inspect.Signature:
    signature = inspect.signature(call)
    globalns = getattr(call, "__globals__", {})
    typed_params = [
        inspect.Parameter(
            name=param.name,
            kind=param.kind,
            default=param.default,
            annotation=get_typed_annotation(param.annotation, globalns),
        )
        for param in signature.parameters.values()
    ]
    return_annotation = signature.return_annotation

    if return_annotation is inspect.Signature.empty:
        return_annotation = None

    globalns = getattr(call, "__globals__", {})
    return_annotation = get_typed_annotation(return_annotation, globalns)
    typed_signature = inspect.Signature(
        typed_params, return_annotation=return_annotation
    )
    return typed_signature


def get_typed_annotation(annotation: t.Any, globalns: dict[str, t.Any]) -> t.Any:
    if isinstance(annotation, str):
        annotation = t.ForwardRef(annotation)
        annotation = evaluate_forwardref(annotation, globalns, globalns)
    return annotation



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass


class TokenError(Exception):
    pass


@dataclass(frozen=True)
class Token(ABC):
    value: str

    @classmethod
    @abstractmethod
    def match(cls, string: str) -> bool: ...  # pragma: no cover


INPUT = t.TypeVar("INPUT")


class Tokenizer[INPUT](ABC):
    @abstractmethod
    def tokenize(self, input: INPUT) -> tuple[Token, ...]: ...  # pragma: no cover



# from dataclasses import dataclass
import typing as t
from dataclasses import dataclass

from griffe import Docstring, DocstringSectionKind, parse_google

from .param_parser import ParamParserAbc
from .parameter import Parameter, Parameters, ReturnValue
from .utils import get_output_info, get_typed_signature

Implicit = t.Literal["input", "param"]


@dataclass(frozen=True)
class OperatorFn:
    """A dataclass to represent an operator function"""

    parameters: Parameters
    output: ReturnValue
    callback: t.Callable[..., t.Any]
    param_parser: ParamParserAbc
    is_delegate: bool = False

    @property
    def short_description(self):
        """Get the short description of the operator"""
        if not self.callback.__doc__:
            return ""
        parsed_docstring = parse_google(Docstring(self.callback.__doc__))

        if parsed_docstring[0].kind != DocstringSectionKind.text:
            return ""

        return parsed_docstring[0].value

    @property
    def long_description(self):
        """Get the long description of the operator"""
        if not self.callback.__doc__:
            return ""
        parsed_docstring = parse_google(Docstring(self.callback.__doc__))

        for section in parsed_docstring:
            if section.title and section.title.lower() == "description":
                return section.value.contents
        return ""

    @property
    def examples(self) -> list[tuple[str, str]]:
        """Get the examples of the operator"""
        if not self.callback.__doc__:
            return []

        parsed_docstring = parse_google(Docstring(self.callback.__doc__))

        for section in parsed_docstring:
            if section.title and (
                section.title.lower() == "operator examples"
                or section.title.lower() == "operator example"
            ):
                example_docstring = parse_google(Docstring(section.value.contents))
                examples: list[tuple[str, str]] = []
                for example in example_docstring:
                    title = "" if example.title is None else example.title
                    contents = (
                        example.value
                        if example.kind == DocstringSectionKind.text
                        else example.value.contents
                    )
                    examples.append((title, contents))
                return examples
        return []

    @classmethod
    def from_def(
        cls,
        func: t.Callable[..., t.Any],
        param_parser: ParamParserAbc,
        implicit: Implicit,
        is_delegate: bool = False,
    ) -> "OperatorFn":
        signature = get_typed_signature(func)
        parameter_list: list[Parameter] = []
        for param in signature.parameters.values():
            parameter_list.append(Parameter.validate(param, implicit))
        return_annotation = signature.return_annotation
        output_info = get_output_info(return_annotation)
        return OperatorFn(
            parameters=Parameters(parameter_list),
            output=ReturnValue.validate(return_annotation, output_info),
            callback=func,
            param_parser=param_parser,
            is_delegate=is_delegate,
        )


class OperatorFns(dict[str, OperatorFn]):
    def __setitem__(self, key: str, value: OperatorFn) -> None:
        assert key not in self, (
            f"Operator '{key}' already exists. Reassignment is not allowed."
        )
        super().__setitem__(key, value)

    def update(self, *arg, **kwds) -> None:
        if kwds or not isinstance(arg[0], OperatorFns):
            raise TypeError(
                f"update() only accept a single positional argument of type {OperatorFns}"
            )
        for key in arg[0].keys():
            assert key not in self, (
                f"Operator '{key}' already exists. Reassignment is not allowed."
            )
        super().update(arg[0])

    def register(
        self,
        *,
        name: str = "",
        param_parser: ParamParserAbc,
        implicit: t.Literal["input", "param"],
        is_delegate: bool = False,
    ) -> t.Callable[..., t.Any]:
        def _decorator(func: t.Callable[..., t.Any]):
            key = name if name else func.__name__
            self[key] = OperatorFn.from_def(
                func,
                param_parser=param_parser,
                implicit=implicit,
                is_delegate=is_delegate,
            )
            return func

        return _decorator



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

from .parameter import Parameters


class ParamParserErrorCtx(t.TypedDict, total=False):
    spos: int
    epos: int
    error: Exception


class ParamParserError(Exception):
    def __init__(
        self,
        message: str,
        ctx: ParamParserErrorCtx = {},
    ) -> None:
        """
        An exception to represent an error in parsing operator arguments

        Args:
            message (str): The error message
            string (str): The string being parsed
            spos (int): The start position of the error substring
            epos (int): The end position of the error substring
        """
        self.message = message
        self.ctx = ctx
        super().__init__(self.message)

    def __str__(self) -> str:
        message = self.message
        if "error" in self.ctx:
            message += f"\n{str(self.ctx['error'])}"
        return message


@dataclass(frozen=True)
class ParamParserAbc(ABC):
    """
    An abstract class to represent an operator argument parser
    """

    @abstractmethod
    def parse_arguments(
        self, string: str, parameters: Parameters
    ) -> tuple[tuple[t.Any, ...], tuple[tuple[str, t.Any], ...]]:
        """
        Parse a parameter string into args and kwds

        Args:
            string (str): The parameter string to parse
            parameters: The list of parameters to parse the string into

        Returns:
            tuple[tuple[str, ...], tuple[KWd, ...]]: A tuple containing a tuple of args and a tuple of kwds

        Raises:
            OprArgParserError: If the parameter string is invalid
        """

    @abstractmethod
    def get_synopsis(self, parameters: Parameters, **kwds: t.Any) -> str:
        """
        Get the synopsis of the operator arguments

        Args:
            parameters: The list of parameters to get the synopsis for
            kwds: Any extra keyword arguments

        Returns:
            str: The synopsis of the operator arguments
        """



import re
import typing as t
from dataclasses import dataclass

from pydantic import ValidationError

from clios.core.param_parser import (
    ParamParserAbc,
    ParamParserError,
)
from clios.core.parameter import Parameters


@dataclass(frozen=True)
class StandardParamParser(ParamParserAbc):
    arg_sep: str = ","
    kw_sep: str = "="
    """
    A basic parameter parser that parses a string into args and kwds

    Split the string by `,` and check if the arg is a keyword argument by checking if it contains `=`

    Example result:
        parse("a,b,c,d=1,e=2") -> (args=("a", "b", "c", "d", "e"), kwds=(KWd("d", "1"), KWd("e", "2")))
    """

    def parse_arguments(
        self,
        string: str,
        parameters: Parameters,
    ) -> tuple[tuple[t.Any, ...], tuple[tuple[str, t.Any], ...]]:
        arg_list = list(string.split(self.arg_sep))

        arg_values: list[t.Any] = []
        kwd_values: dict[str, t.Any] = {}

        positional_arg_iter = parameters.iter_positional_arguments()

        arg_list.reverse()

        spos = 0
        epos = -1
        while len(arg_list) > 0:
            arg = arg_list.pop()
            spos = epos + 1
            epos = spos + len(arg)
            if not arg:
                continue
            kwd = self.get_keyword(arg)
            if kwd is not None:
                k, v = kwd
                if k in kwd_values:
                    raise ParamParserError(
                        f"Duplicate keyword argument `{k}`!",
                        ctx={
                            "spos": spos,
                            "epos": epos,
                        },
                    )
                try:
                    param = parameters.get_keyword_argument(k)
                except KeyError:
                    raise ParamParserError(
                        f"Unknown keyword argument `{k}`!",
                        ctx={
                            "spos": spos,
                            "epos": epos,
                        },
                    )
                try:
                    value = param.build_phase_validator.validate_python(v)
                except ValidationError as e:
                    raise ParamParserError(
                        f"Data validation failed for argument `{k}`!",
                        ctx={
                            "spos": spos,
                            "epos": epos,
                            "error": e,
                        },
                    )
                kwd_values[k] = value
            else:
                if len(kwd_values) > 0:
                    raise ParamParserError(
                        "Positional argument after keyword argument is not allowed!",
                        ctx={
                            "spos": spos,
                            "epos": epos,
                        },
                    )
                try:
                    param = next(positional_arg_iter)
                except StopIteration:
                    raise ParamParserError(
                        f"Too many arguments: expected {len(arg_values)} argument(s)!",
                        ctx={
                            "spos": spos,
                            "epos": epos,
                        },
                    )

                try:
                    value = param.build_phase_validator.validate_python(arg)
                except ValidationError as e:
                    raise ParamParserError(
                        "Data validation failed for argument!",
                        ctx={
                            "spos": spos,
                            "epos": epos,
                            "error": e,
                        },
                    )
                arg_values.append(value)

        if len(arg_values) < parameters.num_required_args:
            raise ParamParserError(
                f"Missing arguments: expected atleast {parameters.num_required_args}, got {len(arg_values)} argument(s)!",
                ctx={
                    "spos": spos,
                    "epos": epos,
                },
            )

        for param_name in parameters.required_keywords.keys():
            if param_name not in kwd_values:
                raise ParamParserError(
                    f"Missing required keyword argument `{param_name}`!",
                    ctx={
                        "spos": spos,
                        "epos": epos,
                    },
                )

        return tuple(arg_values), tuple(kwd_values.items())

    def get_keyword(self, string: str) -> tuple[str, str] | None:
        """split the string into key and value"""

        # Find the first occurence of the kw_sep
        sep_index = string.find(self.kw_sep)

        if sep_index < 1:
            return None

        if not is_valid_variable_name(string[:sep_index]):
            return None

        return string[:sep_index], string[sep_index + 1 :]

    def get_synopsis(self, parameters: Parameters, lsep="", **kwds: t.Any) -> str:
        """get the synopsis of the operator function parameters"""
        required_positional_params: str = ""
        optional_positional_params: str = ""
        required_keyword_params: str = ""
        optional_keyword_params: str = ""

        for param in parameters:
            if param.is_positional_param:
                if param.is_required:
                    required_positional_params += f"{self.arg_sep}{param.name}"
                else:
                    optional_positional_params += f"{self.arg_sep}{param.name}"
            elif param.is_keyword_param:
                if param.is_required:
                    required_keyword_params += (
                        f"{self.arg_sep}{param.name}{self.kw_sep}<val>"
                    )
                else:
                    optional_keyword_params += (
                        f"{self.arg_sep}{param.name}{self.kw_sep}<val>"
                    )
            elif param.is_var_param:
                optional_positional_params += f"{self.arg_sep}*{param.name}"
            elif param.is_var_keyword:
                optional_keyword_params += f"{self.arg_sep}**{param.name}"

        res = required_positional_params
        if optional_positional_params:
            res += f"[{optional_positional_params}]"

        res += required_keyword_params

        if optional_keyword_params:
            res += f"[{optional_keyword_params}]"

        res = res.strip(self.arg_sep)

        if not res:
            return res

        if res[0] == "[":
            res = f"[{lsep}{res[2:]}"
        else:
            res = f"{lsep}{res}"

        return res


def is_valid_variable_name(name: str) -> bool:
    pattern = r"^[a-zA-Z_][a-zA-Z0-9_]*$"
    return re.match(pattern, name) is not None



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

from .operator import RootOperator
from .operator_fn import OperatorFn, OperatorFns


class ParserErrorCtx(t.TypedDict, total=False):
    token_index: int
    num_extra_tokens: int
    unchainable_token_index: int
    error: Exception


class ParserError(Exception):
    def __init__(
        self,
        message: str,
        ctx: ParserErrorCtx = {},
    ) -> None:
        """
        An exception to represent an error in parsing operator arguments

        """
        self.message = message
        self.ctx = ctx
        super().__init__(self.message)

    def __str__(self) -> str:
        message = self.message
        if "error" in self.ctx:
            message += f"\n{str(self.ctx['error'])}"
        return message


@dataclass(frozen=True)
class ParserAbc(ABC):
    @abstractmethod
    def get_operator(
        self, operator_fns: OperatorFns, input: t.Any, **kwds: t.Any
    ) -> RootOperator:
        """
        Parse the tokens and get the tree of operators

        Args:
            operator_fns (OperatorFns): Dictionary of operator functions
            input (t.Any): The input to be parsed
        Returns:
            BaseOperator: The operator
        """

    @abstractmethod
    def is_inline_operator_name(self, name: str) -> bool:
        """
        Check if the string name is an inline operator
        """

    @abstractmethod
    def get_inline_operator_fn(self, name: str, index: int) -> OperatorFn:
        """
        Get an inline operator function

        Args:
            name (str): The operator name
            index (int): The index of the operator

        Returns:
            OperatorFn: The operator function
        """

    @abstractmethod
    def get_synopsis(
        self,
        operator_fn: OperatorFn,
        operator_name: str,
        **kwds: t.Any,
    ) -> str:
        """
        Get the synopsis of an operator

        Args:
            operator_name (str): The operator name

        Returns:
            str: The synopsis
        """
